/*jshint camelcase:false, plusplus:false, proto:true, evil:true, esnext:true, moz:true, unused: vars */
/*global confineAdsrc, setDomainToConfine */

/*
 * Based on the HTML Parser By John Resig (ejohn.org), modified.
 * Original code by Erik Arvidsson, Mozilla Public License
 * http://erik.eae.net/simplehtmlparser/simplehtmlparser.js
 *
 * // Use like so:
 * HTMLParser(htmlString, {
 *     doctype: function(tag) {},
 *     start: function(tag, attrs, unary) {},
 *     end: function(tag) {},
 *     chars: function(text, isLast) {},
 *     comment: function(text) {}
 * });
 */

(function(){
	"use strict";

	// Regular Expressions for parsing tags and attributes
	var startTag = /^<([-A-Za-z0-9_\:]+)((?:\s+[\w-]+(?:\s*=\s*(?:(?:"[^"]*")|(?:'[^']*')|[^>\s]+))?)*)\s*(\/?)>/,
		endTag = /^<\/([-A-Za-z0-9_\:]+)[^>]*>/,
		attr = /([-A-Za-z0-9_]+)(?:\s*=\s*(?:(?:"((?:\\.|[^"])*)")|(?:'((?:\\.|[^'])*)')|([^>\s]+)))?/g,
		doctype = /^<!DOCTYPE\s+[-A-Za-z0-9_]+(?:\s+[-A-Za-z0-9_]+(?:\s+"[^"]*")*)?\s*>/;

	var emptyChars = /^\s*$/;

	var makeMap = function(str){
		var obj = {};
		str.split(",").forEach(function(item) { obj[item] = true; });
		return obj;
	};

	// Attributes that have thejir values filled in disabled="disabled"
	var fillAttrs = makeMap("checked,compact,declare,defer,disabled,ismap,multiple,nohref,noresize,noshade,nowrap,readonly,selected");

	var cdata = makeMap("script,style,textarea");

	var HTMLParser = function( html, handler ) {
		var index, chars, match, savedChars = "", currentTag = "";

		var handleSavedChars = function() {
			if (!emptyChars.test(savedChars)) {
				handler.chars(savedChars);
			}
			savedChars = "";
		};

		var parseStartTag = function( tag, tagName, rest, unary ) {
			// currentTag is only meaningful for cdata tags.
			currentTag = tagName = tagName.toLowerCase();

			unary = !!unary;

			
			var attrs = {};
	
			rest.replace(attr, function(match, name) {
				name = name.toLowerCase();
				var value = arguments[2] ? arguments[2] :
					arguments[3] ? arguments[3] :
					arguments[4] ? arguments[4] :
					fillAttrs[name] ? name : "";
				
				attrs[name] = value;
			});

			handler.start( tagName, attrs, unary );
		};

		while ( html ) {
			chars = true;

			if ( html.indexOf("<!--") === 0 ) {
				index = html.indexOf("-->");

				if ( index >= 0 ) {
					handleSavedChars();
					handler.comment( html.substring( 4, index ) );
					html = html.substring( index + 3 );
					chars = false;
				}
			} else if ( html.indexOf("<!") === 0 && !cdata[currentTag]) {
				match = html.match(doctype);
				if ( match ) {
					handleSavedChars();
					html = html.substring( match[0].length );
					handler.doctype(match[0]); // very crude, we do not even bother parsing or validating
					chars = false;
				}
			} else if ( html.indexOf("</") === 0 ) {
				match = html.match( endTag );
				if ( match && (!cdata[currentTag] || currentTag === match[1].toLowerCase()) ) {
					handleSavedChars();
					html = html.substring( match[0].length );
					handler.end(match[1].toLowerCase());
					currentTag = "";
					chars = false;
				}
			} else if ( html.indexOf("<") === 0 && !cdata[currentTag]) {
				match = html.match( startTag );
				if ( match ) {
					handleSavedChars();
					html = html.substring( match[0].length );
					match[0].replace( startTag, parseStartTag );
					chars = false;
				}
			}

			if ( chars ) {
				index = html.indexOf("<");

				if (index > 0) {
					savedChars += html.substring(0, index);
					html = html.substring(index);
				} else if (index === 0) {
					savedChars += "<";
					html = html.substring(1);
				} else {
					if (!emptyChars.test(savedChars+html))
						handler.chars(savedChars + html, true);
					savedChars = "";
					break;
				}
				
			}
		}

		handleSavedChars();
		
		
	};


	// helpful for debugging
	var logFun = function(n) {
		return function() {
			console.log(n, arguments);
		};
	};
	var loggingHandler = {
		doctype: logFun("doctype"),
		start: logFun("start"),
		end: logFun("end"),
		chars: logFun("chars"),
		comment: logFun("comment")
	};


	var HTMLSerializer = {
		doctype: function(metatag) { return metatag; },
		start: function (tag, attrs, unary) {
			var result = "";
			result += "<" + tag;
			Object.keys(attrs).forEach(function(key) {
				if (attrs[key] === undefined)
					console.log("Fucked up key:", tag, attrs, key);
				result += " " + key + '="' + attrs[key].replace(/"/g, "&quot;") + '"';
			});
			result += (unary ? "/" : "") + ">";
			return result;
		},
		end: function (tag) {
			return "</" + tag + ">";
		},
		chars: function(text) { return text; },
		comment: function(text) {
			return "<!--" + text + "-->";
		},
	};


	window.do_docWrite = function(html, context, mode, w, ads_app) {
		//console.log("Invoke docwrite", arguments);
		ads_app = Cu.waiveXrays(ads_app); // for newer versions of FF, unwrap is not automatic
		var orgHTML = html;
		var result = "";
		var state = {remaining: "", isInlineScript: false, scriptContent: ""}; // preserved accross docwrite calls
		var savedState;
		if (mode.docwrite && (savedState = ads_app.docWriteState[context])) {
			state = savedState;
		}
		var oContext = context;

		html = state.remaining + html;
		state.remaining = "";

		HTMLParser(html, {
			doctype: function(decl) {
				result += decl;
			},
			start: function(tag, attrs, unary) {
				//console.log("start", arguments);
				if (tag === "script" && attrs.context)
					context = attrs.context;				
				if (tag === "script" && !attrs.src && !attrs.privileged && (mode.docwrite || mode.html)) {
					state.isInlineScript = true;
				} else if (tag === "script" && attrs.src && mode.docwrite) {
					setDomainToConfine(attrs.src, context);
				}
				Object.keys(attrs).forEach(function(key) {
					if (key.startsWith("on")){
						if (mode.docwrite || mode.innerHTML)
							attrs[key] = ads_app.confine(attrs[key], context, w.window, true).replace(/\n/g, "");
						else if (mode.html)
							attrs[key] = confineAdsrc(attrs[key], context, w).replace(/\n/g, "");
					}
				});
				if (mode.docwrite)
					attrs["data-ads-app-owner"] = context;
				if (mode.html)
					attrs["data-ads-app-static"] = "true";
				result += HTMLSerializer.start(tag, attrs, unary);
			},
			end: function(tag) {
				//console.log("end", arguments);
				if (state.isInlineScript) {
					state.isInlineScript = false;
				}
				var newScript = state.scriptContent;
				state.scriptContent = "";
				if (newScript) {
					if (mode.docwrite) {
						// FF 32 bug
						// console.log(Object.keys(ads_app), Object.keys(ads_app).length, ads_app, ads_app.confine);
						newScript = ads_app.confine(newScript, context, w, true);
					}
					if (mode.html)
						newScript = confineAdsrc(newScript, context, w);
				}
				result += newScript;
				result += HTMLSerializer.end(tag);
				context = oContext;
			},
			chars: function(text, isFinal) {
				//console.log("chars", arguments);
				if (isFinal && mode.docwrite && text.indexOf("<") > -1) {
					state.remaining = text;
					return;
				}
				if (state.isInlineScript && (mode.docwrite || mode.html))
					state.scriptContent += text;
				else
					result += text;
			},
			comment: function(text) {
				//console.log("comment", arguments);
				if (state.isInlineScript && (mode.docwrite || mode.html))
					state.scriptContent += HTMLSerializer.comment(text);
				else
					result += HTMLSerializer.comment(text);
			}
		});

		if (mode.docwrite)
			ads_app.docWriteState[context] = state;

		// console.log("########");
		// console.log("Input: ", html);
		// console.log("Result: ", result);
		// if (mode === "docwrite")
		// 	console.log("State: ", state);

		//if (mode.docwrite && !state.isInlineScript) // also style, textarea etc should prevent writing this, but yolo
		//	result += "<script>ads_app.processDocWriteNodes(window); document.currentScript.parentNode.removeChild(document.currentScript);</script>";
		
		// TODO: can we save calls to processDocWriteNodes in certain cases?
		if (mode.html){
			dump("==Store HTML src for " + w.url + "\n\n");
			aSession.rewrittenSources[w.url] = result;
		}
		
		return result;
	};

})();


// var html = "<!DOCTYPE aaa public><html><body><script>\n \n<!--puppaaaaaaaaa</script>--></script><style>\n<![CDATA[a: \"</style>\"]]> </style>" +
// "<img src=\"aaa\" b=\"mizzega\" />";
// var html = '<script type="text/javascript">a = 1; console.log(a); f();</script>'
// console.log(html);
// ads_app.HTMLParser(html, ads_app.loggingHandler);